{# Import the theme's layout. #}
{% extends "!layout.html" %}

{% block header %}
<link rel="stylesheet" type="text/css" href="//fonts.googleapis.com/css?family=Vollkorn" /><link rel='stylesheet' href='//fonts.googleapis.com/css?family=Vollkorn' type='text/css' />
<link rel='stylesheet' href='//fonts.googleapis.com/css?family=Vollkorn' type='text/css' />
<link rel='stylesheet' href='//fonts.googleapis.com/css?family=Vollkorn' type='text/css' />
<link rel='stylesheet' href='//fonts.googleapis.com/css?family=Vollkorn' type='text/css' />
<link rel='stylesheet' href='//fonts.googleapis.com/css?family=Vollkorn' type='text/css' />
{% endblock %}

{# Custom CSS overrides #}
{% set bootswatch_css_custom = ['_static/custom.css'] %}
