h1 {
    font-family: Vollkorn;
    font-size: 32px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 32px;
}

h2 {
    font-family: Vollkorn;
    font-size: 28px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 23px;
}

h3 {
    font-family: Vollkorn;
    font-size: 24px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 23px;
}
p, li {
    font-family: Vollkorn;
    font-size: 16px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 23px;
}

li {
    font-size: 12px;
}

blockquote {
    font-family: Vollkorn;
    font-size: 12px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 23px;
}
pre {
    font-family: Vollkorn;
    font-size: 13px;
    font-style: normal;
    font-variant: normal;
    font-weight: 400;
    line-height: 23px;
}
